#include <iostream.h>
#include <algo.h>

#include "simengine.h"

struct DiscreteEventSim & Simengine() {
  static DiscreteEventSim simulator;
  return simulator;
}

DiscreteEventSim::DiscreteEventSim() : 
  eventsExecuted(0),
  currentTime(0),
  finishTime(0),
  terminate(false) { 
}

void DiscreteEventSim::Simulate(double duration) {

  if(duration <= 0.0)
  {
      cerr<<"Simengine: Total simulation time must be greater than 0"<<endl;
      exit(-1);
  }

  finishTime = duration;

  while (currentTime <= finishTime && 
	 (eventQ.empty() == false) &&
	 terminate == false) 
  {
    SimEvent *event = *(eventQ.begin());
    double eventTime = event->EventTime();

    if (eventTime > finishTime)
      break;

    currentTime = eventTime;

#ifdef DEBUG
    cout << "SimTime(" << currentTime << ")" << endl;
#endif

    event->Execute();
    ++eventsExecuted;
    event->Destroy();
  }

  cout << "+++++++++++++++++++++++++++++++++++++++++++++"<<endl
       << "Simulation terminated at simulation time: " << currentTime <<endl
       << "Intended Simulation Duration: " << finishTime <<endl
       << "Total number of events executed during simulation: " << eventsExecuted
       << endl;

}

void DiscreteEventSim::InsertEvent(SimEvent *event) {
  if(event == NULL)
  {
    cout<<"DiscreteEventSim: event must be existent"<<endl;
    exit(0);
  }
  eventQ.insert(event);
}

void DiscreteEventSim::DeleteEvent(SimEvent *event) {
  if (event != NULL) {
    multiset<SimEvent *, CompareEvents>::iterator elem =
      find(eventQ.begin(), eventQ.end(), event);
    
    if (elem != eventQ.end())
      eventQ.erase(elem);
  }
}

inline
double DiscreteEventSim::SimTime() const {
  return currentTime;
}

void DiscreteEventSim::TerminateSimulation() {
  terminate = true;
}



SimEvent::SimEvent(double delay) {
  if(delay<0)
  {
      cout<<"SimEvent: delay has to be >=  0"<<endl;
      exit(0);
  }
  timeOfOccurrence = Simengine().SimTime() + delay;
  Simengine().InsertEvent(this);
}

SimEvent::~SimEvent() {
  Simengine().DeleteEvent(this);
}

void SimEvent::Destroy() {
  delete this;
}

void SimEvent::Execute() {
  cerr << "Event executed at " << Simengine().SimTime() << endl;
}

inline
double SimEvent::EventTime() const {
  return timeOfOccurrence;
}

EndSimulation::EndSimulation(double delay) :
  SimEvent(delay) {
}

EndSimulation::~EndSimulation() {
}

void EndSimulation::Destroy() {
  delete this;
}

void EndSimulation::Execute() {
  Simengine().TerminateSimulation();
}

//******************************************************************

//type:0
DSEvent::DSEvent(void * c_ptr, double delay, void c_mf(void *)):
  SimEvent(delay),
  functype(0),
  s_o_ptr(c_ptr)
{
  myfunct0=c_mf;
}

//type:1
DSEvent::DSEvent(void * c_ptr, double delay, void c_mf(void *, long), long c_lparam):
  SimEvent(delay),
  mylparam1(c_lparam),
  functype(1),
  s_o_ptr(c_ptr)
{
  myfunct1=c_mf;
}

//type:2
DSEvent::DSEvent(void * c_ptr, double delay, void c_mf(void *, long, long), long c_lparam, long c_lparam2):
  SimEvent(delay),
  mylparam1(c_lparam),
  mylparam2(c_lparam2),
  functype(2),
  s_o_ptr(c_ptr)
{
  myfunct2=c_mf;
}

///type:3
DSEvent::DSEvent(void * c_ptr, double delay, void c_mf(void *, double), double c_dparam):
  SimEvent(delay),
  mydparam1(c_dparam),
  functype(3),
  s_o_ptr(c_ptr)
{
  myfunct3=c_mf;
}

//type:4
DSEvent::DSEvent(void * c_ptr, double delay, void c_mf(void *, double, double), double c_dparam, double c_dparam2):
  SimEvent(delay),
  mydparam1(c_dparam),
  mydparam2(c_dparam2),
  functype(4),
  s_o_ptr(c_ptr)
{
  myfunct4=c_mf;
}

//type:5
DSEvent::DSEvent(void * c_ptr, double delay, void c_mf(void *, long, double), long c_lparam, double c_dparam):
  SimEvent(delay),
  mylparam1(c_lparam),
  mydparam1(c_dparam),
  functype(5),
  s_o_ptr(c_ptr)
{
  myfunct5=c_mf;
}

//type:6
DSEvent::DSEvent(void * c_ptr, double delay, void c_mf(void *, long, long, double), long c_lparam, long c_lparam2, double c_dparam):
  SimEvent(delay),
  mylparam1(c_lparam),
  mylparam2(c_lparam2),
  mydparam1(c_dparam),
  functype(6),
  s_o_ptr(c_ptr)
{
  myfunct6=c_mf;
}

//type:7
DSEvent::DSEvent(void * c_ptr, double delay, void c_mf(void *, long, double, double), long c_lparam, double c_dparam, double c_dparam2):
  SimEvent(delay),
  mylparam1(c_lparam),
  mydparam1(c_dparam),
  mydparam2(c_dparam2),
  functype(7),
  s_o_ptr(c_ptr)
{
  myfunct7=c_mf;
}

//type:8
DSEvent::DSEvent(void * c_ptr, double delay, void c_mf(void *, long, long, double, double), long c_lparam, long c_lparam2, double c_dparam, double c_dparam2):
  SimEvent(delay),
  mylparam1(c_lparam),
  mylparam2(c_lparam2),
  mydparam1(c_dparam),
  mydparam2(c_dparam2),
  functype(8),
  s_o_ptr(c_ptr)
{
  myfunct8=c_mf;
}

DSEvent::~DSEvent() {

}

void DSEvent::Destroy() {
    delete this;
}

void DSEvent::Execute() {
  switch(functype)
    {
    case 0:
      myfunct0(s_o_ptr);
      break;
    case 1:
      myfunct1(s_o_ptr,mylparam1);
      break;
    case 2:
      myfunct2(s_o_ptr,mylparam1,mylparam2);
      break;
    case 3:
      myfunct3(s_o_ptr,mydparam1);
      break;
    case 4:
      myfunct4(s_o_ptr,mydparam1,mydparam2);
      break;
    case 5:
      myfunct5(s_o_ptr,mylparam1,mydparam1);
      break;
    case 6:
      myfunct6(s_o_ptr,mylparam1,mylparam2,mydparam1);
      break;
    case 7:
      myfunct7(s_o_ptr,mylparam1,mydparam1,mydparam2);
      break;
    case 8:
      myfunct8(s_o_ptr,mylparam1,mylparam2,mydparam1,mydparam2);
      break;
    default:
      cout<<"Error: undefined event type"<<endl;
      exit(-1);
    }
}

